#!/usr/bin/python
# ################################################################## 
# 
# Copyright 2018 Teradata. All rights reserved.
# TERADATA CONFIDENTIAL AND TRADE SECRET
# 
# Primary Owner: Pankaj Purandare (pankajvinod.purandare@teradata.com)
# Secondary Owner: Mounika Kotha (mounika.kotha@teradata.com)
# 
# Version: 1.2
# Function Version: 1.5
# 
# ################################################################## 

import inspect
from teradataml.common.wrapper_utils import AnalyticsWrapperUtils
from teradataml.common.utils import UtilFuncs
from teradataml.context.context import *
from teradataml.dataframe.dataframe import DataFrame
from teradataml.common.aed_utils import AedUtils
from teradataml.analytics.analytic_query_generator import AnalyticQueryGenerator
from teradataml.common.exceptions import TeradataMlException
from teradataml.common.messages import Messages
from teradataml.common.messagecodes import MessageCodes
from teradataml.common.constants import TeradataConstants
from teradataml.dataframe.dataframe_utils import DataFrameUtils as df_utils
from teradataml.options.display import display
from teradataml.analytics.mle.Arima import Arima

class ArimaPredict:
    
    def __init__(self,
        object = None,
        residual_table = None,
        n_ahead = 1,
        partition_columns = None,
        object_sequence_column = None,
        residual_table_sequence_column = None,
        object_partition_column = None,
        residual_table_partition_column = None,
        object_order_column = None,
        residual_table_order_column = None):
        """
        DESCRIPTION:
            The ArimaPredict function takes as input the ARIMA model produced
            by the function Arima and predicts a specified number of future
            values (time point forecasts) for the modeled sequence.


        PARAMETERS:
            object:
                Required Argument.
                Specifies the name of the teradataml DataFrame that contains the
                model. Such model is generated by Arima function, accessed using
                coefficient (ArimaOuptutObj.coefficient). It can also accept Arima object
                as it's input.

            object_partition_column:
                Required Argument.
                Specifies Partition By columns for object.
                Values to this argument can be provided as list, if multiple columns
                are used for partition.
                Types: str OR list of Strings (str)

            object_order_column:
                Optional Argument.
                Specifies Order By columns for object.
                Values to this argument can be provided as list, if multiple columns
                are used for ordering.
                Types: str OR list of Strings (str)

            residual_table:
                Required Argument.
                Specifies the name of the teradataml DataFrame that contains the
                original input parameters and their residuals. Such object is generated
                by the Arima function and accessed using name residual_table.

            residual_table_partition_column:
                Required Argument.
                Specifies Partition By columns for residual_table.
                Values to this argument can be provided as list, if multiple columns
                are used for partition.
                Types: str OR list of Strings (str)

            residual_table_order_column:
                Required Argument.
                Specifies Order By columns for residual_table.
                Values to this argument can be provided as list, if multiple columns
                are used for ordering.
                Types: str OR list of Strings (str)

            n_ahead:
                Optional Argument.
                Specifies the number of steps to forecast after the end of the time
                series. This value must be a positive integer.
                Default Value: 1
                Types: int

            partition_columns:
                Required Argument.
                Specifies the partition columns that are in model teradataml
                DataFrame and residual table.
                Types: str OR list of Strings (str)

            object_sequence_column:
                Optional Argument.
                Specifies the list of column(s) that uniquely identifies each row of
                the input argument "object". The argument is used to ensure
                deterministic results for functions which produce results that vary
                from run to run.
                Types: str OR list of Strings (str)

            residual_table_sequence_column:
                Optional Argument.
                Specifies the list of column(s) that uniquely identifies each row of
                the input argument "residual_table". The argument is used to ensure
                deterministic results for functions which produce results that vary
                from run to run.
                Types: str OR list of Strings (str)

        RETURNS:
            Instance of ArimaPredict.
            Output teradataml DataFrames can be accessed using attribute
            references, such as ArimaPredictObj.<attribute_name>.
            Output teradataml DataFrame attribute name is:
                result


        RAISES:
            TeradataMlException


        EXAMPLES:
            # Load the data to run the example.
            load_example_data("Arima", "milk_timeseries")

            # Create teradataml DataFrame objects.
            milk_timeseries = DataFrame.from_table("milk_timeseries")

            # Generate arima model using order parameter.
            arima_out = Arima(data = milk_timeseries,
                              timestamp_columns = "period",
                              value_column = "milkpound",
                              order = "3,0,0",
                              include_mean = True)

            # Example 1: Using the generated Arima model to find predictions.
            arima_predict_out1 = ArimaPredict(object = arima_out,
                                                object_partition_column='td_arima_partition_id',
                                                residual_table=arima_out.residual_table,
                                                residual_table_partition_column='td_arima_partition_id',
                                                residual_table_order_column='period',
                                                partition_columns='td_arima_partition_id',
                                                n_ahead=15)

            # Print the result dataframe
            print(arima_predict_out1.result)

            # Example 2: Using coefficient table from a presisted Arima model
            # generated by Arima function.Persist the model tables generated by
            #  Arima function.
            copy_to_sql(arima_out.coefficient, table_name = "arima_coefficients")
            copy_to_sql(arima_out.residual_table, table_name = "arima_residualtable")

            # Create teradataml Dataframe object
            arima_coefficients = DataFrame("arima_coefficients")
            arima_residualtable = DataFrame("arima_residualtable")

            # Find prediction.
            arima_predict_out2 = ArimaPredict(object=arima_coefficients,
                                                 object_partition_column='td_arima_partition_id',
                                                 residual_table=arima_residualtable,
                                                 residual_table_partition_column='td_arima_partition_id',
                                                 residual_table_order_column='period',
                                                 partition_columns='td_arima_partition_id',
                                                 n_ahead=15)

            # Print the result dataframe
            print(arima_predict_out2.result)
        
        """
        self.object  = object 
        self.residual_table  = residual_table 
        self.n_ahead  = n_ahead 
        self.partition_columns  = partition_columns 
        self.object_sequence_column  = object_sequence_column 
        self.residual_table_sequence_column  = residual_table_sequence_column 
        self.object_partition_column  = object_partition_column 
        self.residual_table_partition_column  = residual_table_partition_column 
        self.object_order_column  = object_order_column 
        self.residual_table_order_column  = residual_table_order_column 
        
        # Create TeradataPyWrapperUtils instance which contains validation functions.
        self.__awu = AnalyticsWrapperUtils()
        self.__aed_utils = AedUtils()
        
        # Create argument information matrix to do parameter checking
        self.__arg_info_matrix = []
        self.__arg_info_matrix.append(["object", self.object, False, (DataFrame)])
        self.__arg_info_matrix.append(["object_partition_column", self.object_partition_column, False, (str,list)])
        self.__arg_info_matrix.append(["object_order_column", self.object_order_column, True, (str,list)])
        self.__arg_info_matrix.append(["residual_table", self.residual_table, False, (DataFrame)])
        self.__arg_info_matrix.append(["residual_table_partition_column", self.residual_table_partition_column, False, (str,list)])
        self.__arg_info_matrix.append(["residual_table_order_column", self.residual_table_order_column, False, (str,list)])
        self.__arg_info_matrix.append(["n_ahead", self.n_ahead, False, (int)])
        self.__arg_info_matrix.append(["partition_columns", self.partition_columns, False, (str,list)])
        self.__arg_info_matrix.append(["object_sequence_column", self.object_sequence_column, True, (str,list)])
        self.__arg_info_matrix.append(["residual_table_sequence_column", self.residual_table_sequence_column, True, (str,list)])
        
        if inspect.stack()[1][3] != '_from_model_manager':
            # Perform the function validations
            self.__validate()
            # Generate the ML query
            self.__form_tdml_query()
            # Execute ML query
            self.__execute()
        
    def __validate(self):
        """
        Function to validate sqlmr function arguments, which verifies missing 
        arguments, input argument and table types. Also processes the 
        argument values.
        """
        if isinstance(self.object, Arima):
            self.object = self.object._mlresults[0]
        
        # Make sure that a non-NULL value has been supplied for all mandatory arguments
        self.__awu._validate_missing_required_arguments(self.__arg_info_matrix)
        
        # Make sure that a non-NULL value has been supplied correct type of argument
        self.__awu._validate_argument_types(self.__arg_info_matrix)
        
        # Check to make sure input table types are strings or data frame objects or of valid type.
        self.__awu._validate_input_table_datatype(self.object, "object", Arima)
        self.__awu._validate_input_table_datatype(self.residual_table, "residual_table", None)
        
        # Check whether the input columns passed to the argument are not empty.
        # Also check whether the input columns passed to the argument valid or not.
        self.__awu._validate_input_columns_not_empty(self.partition_columns, "partition_columns")
        self.__awu._validate_dataframe_has_argument_columns(self.partition_columns, "partition_columns", self.object, "object", False)
        
        self.__awu._validate_input_columns_not_empty(self.object_sequence_column, "object_sequence_column")
        self.__awu._validate_dataframe_has_argument_columns(self.object_sequence_column, "object_sequence_column", self.object, "object", False)
        
        self.__awu._validate_input_columns_not_empty(self.residual_table_sequence_column, "residual_table_sequence_column")
        self.__awu._validate_dataframe_has_argument_columns(self.residual_table_sequence_column, "residual_table_sequence_column", self.residual_table, "residual_table", False)
        
        self.__awu._validate_input_columns_not_empty(self.object_partition_column, "object_partition_column")
        self.__awu._validate_dataframe_has_argument_columns(self.object_partition_column, "object_partition_column", self.object, "object", False)
        
        self.__awu._validate_input_columns_not_empty(self.residual_table_partition_column, "residual_table_partition_column")
        self.__awu._validate_dataframe_has_argument_columns(self.residual_table_partition_column, "residual_table_partition_column", self.residual_table, "residual_table", False)
        
        self.__awu._validate_input_columns_not_empty(self.object_order_column, "object_order_column")
        self.__awu._validate_dataframe_has_argument_columns(self.object_order_column, "object_order_column", self.object, "object", False)
        
        self.__awu._validate_input_columns_not_empty(self.residual_table_order_column, "residual_table_order_column")
        self.__awu._validate_dataframe_has_argument_columns(self.residual_table_order_column, "residual_table_order_column", self.residual_table, "residual_table", False)
        
        
    def __form_tdml_query(self):
        """
        Function to generate the analytical function queries. The function defines 
        variables and list of arguments required to form the query.
        """
        
        # Output table arguments list
        self.__func_output_args_sql_names = []
        self.__func_output_args = []
        
        # Generate lists for rest of the function arguments
        self.__func_other_arg_sql_names = []
        self.__func_other_args = []
        self.__func_other_arg_json_datatypes = []
        
        self.__func_other_arg_sql_names.append("PartitionColumns")
        self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(UtilFuncs._teradata_quote_arg(self.partition_columns,"\""),"'"))
        self.__func_other_arg_json_datatypes.append("COLUMNS")
        
        self.__func_other_arg_sql_names.append("StepAhead")
        self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(self.n_ahead,"'"))
        self.__func_other_arg_json_datatypes.append("INTEGER")
        
        # Generate lists for rest of the function arguments
        sequence_input_by_list = []
        if self.object_sequence_column is not None:
            sequence_input_by_list.append("ModelTable:" + UtilFuncs._teradata_collapse_arglist(self.object_sequence_column,""))
        
        if self.residual_table_sequence_column is not None:
            sequence_input_by_list.append("ResidualTable:" + UtilFuncs._teradata_collapse_arglist(self.residual_table_sequence_column,""))
        
        if len(sequence_input_by_list) > 0:
            self.__func_other_arg_sql_names.append("SequenceInputBy")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(sequence_input_by_list,"'"))
            self.__func_other_arg_json_datatypes.append("STRING")
        
        
        # Declare empty lists to hold input table information.
        self.__func_input_arg_sql_names = []
        self.__func_input_table_view_query = []
        self.__func_input_dataframe_type = []
        self.__func_input_distribution = []
        self.__func_input_partition_by_cols = []
        self.__func_input_order_by_cols = []
        
        # Process object
        self.object_partition_column = UtilFuncs._teradata_collapse_arglist(self.object_partition_column,"\"")
        self.__table_ref = self.__awu._teradata_on_clause_from_dataframe(self.object, False)
        self.__func_input_distribution.append("FACT")
        self.__func_input_arg_sql_names.append("ModelTable")
        self.__func_input_table_view_query.append(self.__table_ref["ref"])
        self.__func_input_dataframe_type.append(self.__table_ref["ref_type"])
        self.__func_input_partition_by_cols.append(self.object_partition_column)
        self.__func_input_order_by_cols.append(UtilFuncs._teradata_collapse_arglist(self.object_order_column,"\""))
        
        # Process residual_table
        self.residual_table_partition_column = UtilFuncs._teradata_collapse_arglist(self.residual_table_partition_column,"\"")
        self.__table_ref = self.__awu._teradata_on_clause_from_dataframe(self.residual_table, False)
        self.__func_input_distribution.append("FACT")
        self.__func_input_arg_sql_names.append("ResidualTable")
        self.__func_input_table_view_query.append(self.__table_ref["ref"])
        self.__func_input_dataframe_type.append(self.__table_ref["ref_type"])
        self.__func_input_partition_by_cols.append(self.residual_table_partition_column)
        self.__func_input_order_by_cols.append(UtilFuncs._teradata_collapse_arglist(self.residual_table_order_column,"\""))
        
        function_name = "ARIMAPredict"
        # Create instance to generate SQLMR.
        aqg_obj = AnalyticQueryGenerator(function_name 
                ,self.__func_input_arg_sql_names 
                ,self.__func_input_table_view_query 
                ,self.__func_input_dataframe_type 
                ,self.__func_input_distribution 
                ,self.__func_input_partition_by_cols 
                ,self.__func_input_order_by_cols 
                ,self.__func_other_arg_sql_names 
                ,self.__func_other_args 
                ,self.__func_other_arg_json_datatypes 
                ,self.__func_output_args_sql_names 
                ,self.__func_output_args 
                ,engine = "ENGINE_ML")
        # Invoke call to SQL-MR generation.
        self.sqlmr_query = aqg_obj._gen_sqlmr_select_stmt_sql()
        
        # Print SQL-MR query if requested to do so.
        if display.print_sqlmr_query:
            print(self.sqlmr_query)
        
    def __execute(self):
        """
        Function to execute SQL-MR queries. 
        Create DataFrames for the required SQL-MR outputs.
        """
        # Generate STDOUT table name and add it to the output table list.
        sqlmr_stdout_temp_tablename = UtilFuncs._generate_temp_table_name(prefix = "td_sqlmr_out_", use_default_database = True, gc_on_quit = True, quote=False)
        try:
            UtilFuncs._create_view(sqlmr_stdout_temp_tablename, self.sqlmr_query)
        except Exception as emsg:
            raise TeradataMlException(Messages.get_message(MessageCodes.TDMLDF_EXEC_SQL_FAILED, str(emsg)), MessageCodes.TDMLDF_EXEC_SQL_FAILED)
        
        # Update output table data frames.
        self._mlresults = []
        self.result = self.__awu._create_data_set_object(df_input=UtilFuncs._extract_table_name(sqlmr_stdout_temp_tablename), source_type="table", database_name=UtilFuncs._extract_db_name(sqlmr_stdout_temp_tablename))
        self._mlresults.append(self.result)
        
    @classmethod
    def _from_model_manager(cls,
        result = None,
        **kwargs):
        """
        Classmethod which will be used by Model Manager, to instantiate this wrapper class.
        """
        kwargs.pop("result", None)
        
        # Let's create an object of this class.
        obj = cls(**kwargs)
        obj.result  = result 
        
        # Update output table data frames.
        obj._mlresults = []
        obj.result = obj.__awu._create_data_set_object(df_input=UtilFuncs._extract_table_name(obj.result), source_type="table", database_name=UtilFuncs._extract_db_name(obj.result))
        obj._mlresults.append(obj.result)
        return obj
        
    def __repr__(self):
        """
        Returns the string representation for a ArimaPredict class instance.
        """
        repr_string="############ STDOUT Output ############"
        repr_string = "{}\n\n{}".format(repr_string,self.result)
        return repr_string
        
