#!/usr/bin/python
# ################################################################## 
# 
# Copyright 2018 Teradata. All rights reserved.
# TERADATA CONFIDENTIAL AND TRADE SECRET
# 
# Primary Owner: Bhavana N (bhavana.n@teradata.com)
# Secondary Owner: Pankaj Purandare (pankajvinod.purandare@teradata.com)
# 
# Version: 1.2
# Function Version: 1.12
# 
# ################################################################## 

import inspect
from teradataml.common.wrapper_utils import AnalyticsWrapperUtils
from teradataml.common.utils import UtilFuncs
from teradataml.context.context import *
from teradataml.dataframe.dataframe import DataFrame
from teradataml.common.aed_utils import AedUtils
from teradataml.analytics.analytic_query_generator import AnalyticQueryGenerator
from teradataml.common.exceptions import TeradataMlException
from teradataml.common.messages import Messages
from teradataml.common.messagecodes import MessageCodes
from teradataml.common.constants import TeradataConstants
from teradataml.dataframe.dataframe_utils import DataFrameUtils as df_utils
from teradataml.options.display import display
from teradataml.analytics.mle.GLM import GLM

class GLMPredict:
    
    def __init__(self,
        modeldata = None,
        newdata = None,
        terms = None,
        family = None,
        linkfunction = "CANONICAL",
        output_response_probdist = False,
        output_responses = None,
        newdata_sequence_column = None,
        modeldata_sequence_column = None,
        newdata_order_column = None,
        modeldata_order_column = None):
        """
        DESCRIPTION:
            The GLMPredict function uses the model generated by the function GLM
            to perform generalized linear model prediction on new input data.

            Note: This function is available only when teradataml is connected to
                  Vantage 1.1 or later versions.

        PARAMETERS:
            modeldata:
                Required Argument.
                Specifies the teradataml DataFrame containing the model data.

            modeldata_order_column:
                Optional Argument.
                Specifies Order By columns for modeldata.
                Values to this argument can be provided as a list, if multiple columns
                are used for ordering.
                Types: str OR list of Strings (str)

            newdata:
                Required Argument.
                Specifies the teradataml DataFrame containing the input data.

            newdata_order_column:
                Optional Argument.
                Specifies Order By columns for newdata.
                Values to this argument can be provided as a list, if multiple columns
                are used for ordering.
                Types: str OR list of Strings (str)

            terms:
                Optional Argument.
                Specifies the names of input teradataml DataFrame columns to copy to
                the output teradataml DataFrame.
                Types: str OR list of Strings (str)

            family:
                Optional Argument.
                Specifies the distribution exponential family. The default value is
                read from model teradataml DataFrame. If you specify this argument, you must give it
                the same value that you used for the "family" argument of the function
                when you generated the model teradataml DataFrame.
                Permitted Values: LOGISTIC, BINOMIAL, POISSON, GAUSSIAN, GAMMA,
                INVERSE_GAUSSIAN, NEGATIVE_BINOMIAL
                Types: str

            linkfunction:
                Optional Argument.
                Specifies the canonical link functions and the link
                functions that are allowed for each exponential family.
                Note: Use the same value that you used for the "link" argument
                      of the function when you generated the model teradataml DataFrame.
                Default Value: "CANONICAL"
                Permitted Values: CANONICAL, IDENTITY, INVERSE, LOG,
                COMPLEMENTARY_LOG_LOG, SQUARE_ROOT, INVERSE_MU_SQUARED, LOGIT,
                PROBIT, CAUCHIT
                Types: str
            
            output_response_probdist:
                Optional Argument.
                Specifies whether to output probabilities.
                Note: "output_response_probdist" argument support is only available 
                      when teradataml is connected to Vantage 1.1.1 or later.
                Default Value: False
                Types: bool
            
            output_responses:
                Optional Argument.
                Specifies responses to output probability.
                This argument can only be used when output_response_probdist is set to TRUE.
                Note: "output_responses" argument support is only available when 
                      teradataml is connected to Vantage 1.1.1 or later.
                Permitted Values: 0, 1
                Types: str OR list of Strings (str)

            newdata_sequence_column:
                Optional Argument.
                Specifies the list of column(s) that uniquely identifies each row of
                the input argument "newdata". The argument is used to ensure
                deterministic results for functions which produce results that vary
                from run to run.
                Types: str OR list of Strings (str)

            modeldata_sequence_column:
                Optional Argument.
                Specifies the list of column(s) that uniquely identifies each row of
                the input argument "modeldata". The argument is used to ensure
                deterministic results for functions which produce results that vary
                from run to run.
                Types: str OR list of Strings (str)

        RETURNS:
            Instance of GLMPredict.
            Output teradataml DataFrames can be accessed using attribute
            references, such as GLMPredictObj.<attribute_name>.
            Output teradataml DataFrame attribute name is:
                result

        RAISES:
            TeradataMlException


        EXAMPLES:
            # Load the data to run the example.
            load_example_data("GLMPredict", ["admissions_test","admissions_train","housing_test","housing_train"])

            # Create teradataml DataFrame objects.
            admissions_test = DataFrame.from_table("admissions_test")
            admissions_train = DataFrame.from_table("admissions_train")
            housing_test = DataFrame.from_table("housing_test")
            housing_train = DataFrame.from_table("housing_train")

            # Example 1 -
            # Generate a model based on train data "admissions_train"
            glm_out = GLM(formula = "admitted ~ stats + masters + gpa + programming",
                         family = "LOGISTIC",
                         linkfunction = "LOGIT",
                         data = admissions_train,
                         weights = "1",
                         threshold = 0.01,
                         maxit = 25,
                         step = False,
                         intercept = True
                         )

            # Use the generated model to predict the 'admissions' on the test data
            # admissions_test by using generated model by GLM.
            glm_predict_out1 = GLMPredict(modeldata = glm_out,
                                          newdata = admissions_test,
                                          terms = ["id","masters","gpa","stats","programming","admitted"],
                                          family = "LOGISTIC",
                                          linkfunction = "LOGIT",
                                          output_response_probdist=True
                                          )

            # Print the result DataFrame.
            print(glm_predict_out1.result)

            # Example 2 - Generate a model based on train data "housing_train"
            glm_out_hs = GLM(formula = "price  ~ recroom  + lotsize  + stories  + garagepl + gashw \
                                       + bedrooms  + driveway  + airco  + homestyle + bathrms  + fullbase \
                                       + prefarea",
                              family = "GAUSSIAN",
                              linkfunction = "IDENTITY",
                              data = housing_train,
                              weights = "1",
                              threshold = 0.01,
                              maxit = 25,
                              step = False,
                              intercept = True
                              )

            # Use the generated model to predict the 'price' on the test data
            # housing_test by using generated model by GLM.
            glm_predict_out2 = GLMPredict(modeldata = glm_out_hs.coefficients,
                                          newdata = housing_test,
                                          terms = ["sn", "price"],
                                          family = "GAUSSIAN",
                                          linkfunction = "CANONICAL"
                                          )

            # Print the result DataFrame.
            print(glm_predict_out2)
        """
        self.modeldata  = modeldata 
        self.newdata  = newdata 
        self.terms  = terms 
        self.family  = family 
        self.linkfunction  = linkfunction 
        self.output_response_probdist  = output_response_probdist 
        self.output_responses  = output_responses 
        self.newdata_sequence_column  = newdata_sequence_column 
        self.modeldata_sequence_column  = modeldata_sequence_column 
        self.newdata_order_column  = newdata_order_column 
        self.modeldata_order_column  = modeldata_order_column 
        
        # Create TeradataPyWrapperUtils instance which contains validation functions.
        self.__awu = AnalyticsWrapperUtils()
        self.__aed_utils = AedUtils()
        
        # Create argument information matrix to do parameter checking
        self.__arg_info_matrix = []
        self.__arg_info_matrix.append(["modeldata", self.modeldata, False, (DataFrame)])
        self.__arg_info_matrix.append(["modeldata_order_column", self.modeldata_order_column, True, (str,list)])
        self.__arg_info_matrix.append(["newdata", self.newdata, False, (DataFrame)])
        self.__arg_info_matrix.append(["newdata_order_column", self.newdata_order_column, True, (str,list)])
        self.__arg_info_matrix.append(["terms", self.terms, True, (str,list)])
        self.__arg_info_matrix.append(["family", self.family, True, (str)])
        self.__arg_info_matrix.append(["linkfunction", self.linkfunction, True, (str)])
        self.__arg_info_matrix.append(["output_response_probdist", self.output_response_probdist, True, (bool)])
        self.__arg_info_matrix.append(["output_responses", self.output_responses, True, (str,list)])
        self.__arg_info_matrix.append(["newdata_sequence_column", self.newdata_sequence_column, True, (str,list)])
        self.__arg_info_matrix.append(["modeldata_sequence_column", self.modeldata_sequence_column, True, (str,list)])
        
        if inspect.stack()[1][3] != '_from_model_manager':
            # Perform the function validations
            self.__validate()
            # Generate the ML query
            self.__form_tdml_query()
            # Execute ML query
            self.__execute()
        
    def __validate(self):
        """
        Function to validate sqlmr function arguments, which verifies missing 
        arguments, input argument and table types. Also processes the 
        argument values.
        """
        if isinstance(self.modeldata, GLM):
            self.modeldata = self.modeldata._mlresults[0]
        
        # To use output_responses, output_response_probdist must be set to True
        if self.output_response_probdist is False and self.output_responses is not None:
            raise TeradataMlException(Messages.get_message(MessageCodes.DEPENDENT_ARG_MISSING,
                                                           'output_response_probdist=True',
                                                           'output_responses'),
                                      MessageCodes.DEPENDENT_ARG_MISSING)
        
        # Make sure that a non-NULL value has been supplied for all mandatory arguments
        self.__awu._validate_missing_required_arguments(self.__arg_info_matrix)
        
        # Make sure that a non-NULL value has been supplied correct type of argument
        self.__awu._validate_argument_types(self.__arg_info_matrix)
        
        # Check to make sure input table types are strings or data frame objects or of valid type.
        self.__awu._validate_input_table_datatype(self.newdata, "newdata", None)
        self.__awu._validate_input_table_datatype(self.modeldata, "modeldata", GLM)
        
        # Check for permitted values
        family_permitted_values = ["LOGISTIC", "BINOMIAL", "POISSON", "GAUSSIAN", "GAMMA", "INVERSE_GAUSSIAN", "NEGATIVE_BINOMIAL"]
        self.__awu._validate_permitted_values(self.family, family_permitted_values, "family")
        
        linkfunction_permitted_values = ["CANONICAL", "IDENTITY", "INVERSE", "LOG", "COMPLEMENTARY_LOG_LOG", "SQUARE_ROOT", "INVERSE_MU_SQUARED", "LOGIT", "PROBIT", "CAUCHIT"]
        self.__awu._validate_permitted_values(self.linkfunction, linkfunction_permitted_values, "linkfunction")
        
        output_responses_permitted_values = ["0", "1"]
        self.__awu._validate_permitted_values(self.output_responses, output_responses_permitted_values, "output_responses")
        
        # Check whether the input columns passed to the argument are not empty.
        # Also check whether the input columns passed to the argument valid or not.
        self.__awu._validate_input_columns_not_empty(self.terms, "terms")
        self.__awu._validate_dataframe_has_argument_columns(self.terms, "terms", self.newdata, "newdata", False)
        
        self.__awu._validate_input_columns_not_empty(self.newdata_sequence_column, "newdata_sequence_column")
        self.__awu._validate_dataframe_has_argument_columns(self.newdata_sequence_column, "newdata_sequence_column", self.newdata, "newdata", False)
        
        self.__awu._validate_input_columns_not_empty(self.modeldata_sequence_column, "modeldata_sequence_column")
        self.__awu._validate_dataframe_has_argument_columns(self.modeldata_sequence_column, "modeldata_sequence_column", self.modeldata, "modeldata", False)
        
        self.__awu._validate_input_columns_not_empty(self.newdata_order_column, "newdata_order_column")
        self.__awu._validate_dataframe_has_argument_columns(self.newdata_order_column, "newdata_order_column", self.newdata, "newdata", False)
        
        self.__awu._validate_input_columns_not_empty(self.modeldata_order_column, "modeldata_order_column")
        self.__awu._validate_dataframe_has_argument_columns(self.modeldata_order_column, "modeldata_order_column", self.modeldata, "modeldata", False)
        
        
    def __form_tdml_query(self):
        """
        Function to generate the analytical function queries. The function defines 
        variables and list of arguments required to form the query.
        """
        
        # Output table arguments list
        self.__func_output_args_sql_names = []
        self.__func_output_args = []
        
        # Generate lists for rest of the function arguments
        self.__func_other_arg_sql_names = []
        self.__func_other_args = []
        self.__func_other_arg_json_datatypes = []
        
        if self.terms is not None:
            self.__func_other_arg_sql_names.append("Accumulate")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(UtilFuncs._teradata_quote_arg(self.terms,"\""),"'"))
            self.__func_other_arg_json_datatypes.append("COLUMNS")
        
        if self.family is not None:
            self.__func_other_arg_sql_names.append("Family")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(self.family,"'"))
            self.__func_other_arg_json_datatypes.append("STRING")
        
        if self.linkfunction is not None and self.linkfunction != "CANONICAL":
            self.__func_other_arg_sql_names.append("LinkFunction")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(self.linkfunction,"'"))
            self.__func_other_arg_json_datatypes.append("STRING")
        
        if self.output_response_probdist is not None and self.output_response_probdist != False:
            self.__func_other_arg_sql_names.append("OutputProb")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(self.output_response_probdist,"'"))
            self.__func_other_arg_json_datatypes.append("BOOLEAN")
        
        if self.output_responses is not None:
            self.__func_other_arg_sql_names.append("Responses")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(self.output_responses,"'"))
            self.__func_other_arg_json_datatypes.append("STRING")
        
        # Generate lists for rest of the function arguments
        sequence_input_by_list = []
        if self.newdata_sequence_column is not None:
            sequence_input_by_list.append("input:" + UtilFuncs._teradata_collapse_arglist(self.newdata_sequence_column,""))
        
        if self.modeldata_sequence_column is not None:
            sequence_input_by_list.append("model:" + UtilFuncs._teradata_collapse_arglist(self.modeldata_sequence_column,""))
        
        if len(sequence_input_by_list) > 0:
            self.__func_other_arg_sql_names.append("SequenceInputBy")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(sequence_input_by_list,"'"))
            self.__func_other_arg_json_datatypes.append("STRING")
        
        
        # Declare empty lists to hold input table information.
        self.__func_input_arg_sql_names = []
        self.__func_input_table_view_query = []
        self.__func_input_dataframe_type = []
        self.__func_input_distribution = []
        self.__func_input_partition_by_cols = []
        self.__func_input_order_by_cols = []
        
        # Process newdata
        self.__table_ref = self.__awu._teradata_on_clause_from_dataframe(self.newdata, False)
        self.__func_input_distribution.append("FACT")
        self.__func_input_arg_sql_names.append("input")
        self.__func_input_table_view_query.append(self.__table_ref["ref"])
        self.__func_input_dataframe_type.append(self.__table_ref["ref_type"])
        self.__func_input_partition_by_cols.append("ANY")
        self.__func_input_order_by_cols.append(UtilFuncs._teradata_collapse_arglist(self.newdata_order_column,"\""))
        
        # Process modeldata
        self.__table_ref = self.__awu._teradata_on_clause_from_dataframe(self.modeldata, False)
        self.__func_input_distribution.append("DIMENSION")
        self.__func_input_arg_sql_names.append("model")
        self.__func_input_table_view_query.append(self.__table_ref["ref"])
        self.__func_input_dataframe_type.append(self.__table_ref["ref_type"])
        self.__func_input_partition_by_cols.append("NA_character_")
        self.__func_input_order_by_cols.append(UtilFuncs._teradata_collapse_arglist(self.modeldata_order_column, "\""))
        
        function_name = "GLMPredict"
        # Create instance to generate SQLMR.
        aqg_obj = AnalyticQueryGenerator(function_name 
                ,self.__func_input_arg_sql_names 
                ,self.__func_input_table_view_query 
                ,self.__func_input_dataframe_type 
                ,self.__func_input_distribution 
                ,self.__func_input_partition_by_cols 
                ,self.__func_input_order_by_cols 
                ,self.__func_other_arg_sql_names 
                ,self.__func_other_args 
                ,self.__func_other_arg_json_datatypes 
                ,self.__func_output_args_sql_names 
                ,self.__func_output_args 
                ,engine = "ENGINE_ML")
        # Invoke call to SQL-MR generation.
        self.sqlmr_query = aqg_obj._gen_sqlmr_select_stmt_sql()
        
        # Print SQL-MR query if requested to do so.
        if display.print_sqlmr_query:
            print(self.sqlmr_query)
        
    def __execute(self):
        """
        Function to execute SQL-MR queries. 
        Create DataFrames for the required SQL-MR outputs.
        """
        # Generate STDOUT table name and add it to the output table list.
        sqlmr_stdout_temp_tablename = UtilFuncs._generate_temp_table_name(prefix = "td_sqlmr_out_", use_default_database = True, gc_on_quit = True, quote=False)
        try:
            UtilFuncs._create_view(sqlmr_stdout_temp_tablename, self.sqlmr_query)
        except Exception as emsg:
            raise TeradataMlException(Messages.get_message(MessageCodes.TDMLDF_EXEC_SQL_FAILED, str(emsg)), MessageCodes.TDMLDF_EXEC_SQL_FAILED)
        
        # Update output table data frames.
        self._mlresults = []
        self.result = self.__awu._create_data_set_object(df_input=UtilFuncs._extract_table_name(sqlmr_stdout_temp_tablename), source_type="table", database_name=UtilFuncs._extract_db_name(sqlmr_stdout_temp_tablename))
        self._mlresults.append(self.result)
        
    @classmethod
    def _from_model_manager(cls,
        result = None,
        **kwargs):
        """
        Classmethod which will be used by Model Manager, to instantiate this wrapper class.
        """
        kwargs.pop("result", None)
        
        # Let's create an object of this class.
        obj = cls(**kwargs)
        obj.result  = result 
        
        # Update output table data frames.
        obj._mlresults = []
        obj.result = obj.__awu._create_data_set_object(df_input=UtilFuncs._extract_table_name(obj.result), source_type="table", database_name=UtilFuncs._extract_db_name(obj.result))
        obj._mlresults.append(obj.result)
        return obj
        
    def __repr__(self):
        """
        Returns the string representation for a GLMPredict class instance.
        """
        repr_string="############ STDOUT Output ############"
        repr_string = "{}\n\n{}".format(repr_string,self.result)
        return repr_string
        
